# -*- coding: utf-8 -*-

import sage.all


from .numperiods.family import Family
from .numperiods.integerRelations import IntegerRelations
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.qqbar import QQbar
from sage.functions.other import factorial
from sage.functions.other import floor
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix

from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.schemes.toric.weierstrass import WeierstrassForm
from sage.misc.flatten import flatten

from sage.modules.free_quadratic_module_integer_symmetric import IntegralLattice

from .voronoi import FundamentalGroupVoronoi
from .integrator import Integrator
from .util import Util
from .context import Context
from .hypersurface import Hypersurface
from .ellipticSingularity import EllipticSingularities

import logging
import time

logger = logging.getLogger(__name__)


class EllipticSurface(object):
    def __init__(self, P, basepoint=None, fibration=None, **kwds) -> None:
        """P, a homogeneous polynomial defining an.

        This class aims at computing an effective basis of the homology group H_n(X), 
        given as lifts of paths through a Lefschetz fibration.
        """
        
        self.ctx = Context(**kwds)
        
        # maybe put P in Weierstrass form ?
        # flatPolRing = PolynomialRing(QQ, ['a','b','c','t'])
        # [a,b,c,t] = flatPolRing.gens()
        # weierstrass_coefs = WeierstrassForm(P(t,a,b,c), [a,b,c])
        # Qt = PolynomialRing(QQ, 't')
        # t= P.parent().gens()[0]
        # weierstrass_coefs =  [c(t=t) for c in weierstrass_coefs]
        # t= P.parent().gens()[0]
        # X,Y,Z = P.base_ring().gens()
        # P = -Y**2*Z + X**3 + weierstrass_coefs[0]*X*Z**2 + weierstrass_coefs[1]*Z**3

        self._P = P
        self._family = Family(self.P)
        self._fibration = fibration

        if basepoint!= None: # it is useful to be able to specify the basepoint to avoid being stuck in arithmetic computations if critical values have very large modulus
            assert basepoint not in self.critical_values, "basepoint is not regular"
            self._basepoint=basepoint
        if not self.ctx.debug:
            fg = self.fundamental_group # this allows reordering the critical points straight away and prevents shenanigans. There should be a better way to do this
    
    
    @property
    def intersection_product(self):
        if not hasattr(self,'_intersection_product'):
            self._intersection_product=self._compute_intersection_product()
        return self._intersection_product

    @property
    def primary_periods(self):
        if not hasattr(self, '_primary_periods'):
            homology_mat = matrix(self.extensions).transpose()
            integrated_thimbles =  matrix(self.integrated_thimbles)
            self._primary_periods = integrated_thimbles*homology_mat

        return self._primary_periods
    
    @property
    def periods_matrix(self):
        if not hasattr(self, '_periods_matrix'):
            periods_tot = block_matrix([[self.primary_periods, zero_matrix(len(self.holomorphic_forms), len(flatten(self.singular_components))+2)]])
            self._periods_matrix = periods_tot*matrix(self.primary_lattice).transpose()**-1
        return self._periods_matrix

    @property
    def P(self):
        return self._P

    @property
    def picard_fuchs_equations(self):
        if not hasattr(self,'_picard_fuchs_equations'):
            self._picard_fuchs_equations = [self._family.picard_fuchs_equation(vector([w,0])) for w in self.holomorphic_forms]
        return self._picard_fuchs_equations
    
    @property
    def family(self):
        return self._family
    
    @property
    def discriminant(self):
        if not hasattr(self,'_discriminant'):
            flatPolRing = PolynomialRing(QQ, ['a','b','c','t'])
            [a,b,c,t] = flatPolRing.gens()
            weierstrass_coefs = WeierstrassForm(self.P(t,a,b,c), [a,b,c])
            Qt = PolynomialRing(QQ, 't')
            t = Qt.gens()[0]
            weierstrass_coefs =  [c(t=t) for c in weierstrass_coefs]
            self._discriminant=Qt(4*weierstrass_coefs[0](t=t)**3 + 27*weierstrass_coefs[1](t=t)**2)
        return self._discriminant
    
    @property
    def critical_values(self):
        if not hasattr(self,'_critical_values'):
            self._critical_values=self.discriminant.roots(QQbar, multiplicities=False)
        return self._critical_values
    
    @property
    def monodromy_matrices(self):
        if not hasattr(self, '_monodromy_matrices'):
            n = len(self.fiber.extensions) 
            
            cyclic_form = self.cyclic_form
            w = cyclic_form[0]*self.P + cyclic_form[1]*self.family.coho1.basis()[1]

            integration_correction = diagonal_matrix([1/ZZ(factorial(k)) for k in range(n+1)])
            derivatives_at_basepoint = self.derivatives_values_at_basepoint(w)
            cohomology_fiber_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fiber.cohomology], self.basepoint)

            initial_conditions = integration_correction * derivatives_at_basepoint * cohomology_fiber_to_family.inverse()
            initial_conditions = initial_conditions.submatrix(1,0)

            cohomology_monodromies = [initial_conditions**(-1)*M.submatrix(1,1)*initial_conditions for M in self.cyclic_transition_matrices]

            Ms = [(self.fiber.period_matrix**(-1)*M*self.fiber.period_matrix) for M in cohomology_monodromies]
            if not self.ctx.debug:
                Ms = [M.change_ring(ZZ) for M in Ms]
            
            Mtot=identity_matrix(2)
            for M in Ms:
                Mtot=M*Mtot
            if Mtot!=identity_matrix(2):
                self._critical_values = self.critical_values+["infinity"]
                transition_matrix_infinity = 1
                for M in self.cyclic_transition_matrices:
                    transition_matrix_infinity = M*transition_matrix_infinity
                self._cyclic_transition_matrices += [transition_matrix_infinity**(-1)]
                Ms += [(Mtot**-1).change_ring(ZZ)]
                pathtot=[]
                for path in self.paths:
                    pathtot=pathtot+path
                self._paths+=[list(reversed(Util.simplify_path(pathtot)))]
            
            self._monodromy_matrices = Ms
        return self._monodromy_matrices
    
    @property
    def thimbles_confluence(self):
        if not hasattr(self, '_thimbles_confluence'):
            blocks =[]
            for i, pcs in enumerate(self.permuting_cycles):
                decompositions = []
                for p in pcs:
                    decomposition = []
                    for M, v in zip(self.monodromy_matrices_morsification[i], self.vanishing_cycles_morsification[i]):
                        decomposition += [(M-1)*p/v]
                        p = M*p
                    decompositions+=[decomposition]
                blocks+= [matrix(decompositions)]
            self._thimbles_confluence = block_diagonal_matrix(blocks).change_ring(ZZ)
        return self._thimbles_confluence
    
    def morsify(self, v):
        """Given an extension of the surface, yields its description as an extension of the morsification."""
        return v*self.thimbles_confluence

    @property
    def vanishing_cycles_morsification(self):
        if not hasattr(self, '_vanishing_cycles_morsification'):
            self._vanishing_cycles_morsification = [[(M-1).transpose().image().gens()[0] for M in Ms] for Ms in self.monodromy_matrices_morsification]
        return self._vanishing_cycles_morsification
    
    @property
    def singular_components(self):
        if not hasattr(self, '_singular_components'):
            # fullmat = block_matrix([[self.homology], [self.infinity_loops*self.thimbles_confluence]])
            ranktot = 0
            rankmax = sum([len(Ms) for Ms in self.monodromy_matrices_morsification])
            sing_comps = []
            for M in self.vanishing_cycles_morsification:
                M = matrix(M)
                rank = M.dimensions()[0]
                sing_comps += [[vector([0]*ranktot+list(v) + [0]*(rankmax-ranktot-rank)) for v in M.kernel().gens()]]
                ranktot+=rank
            self._singular_components = sing_comps # [[fullmat.solve_left(component)[:-self.infinity_loops.nrows()] for component in components] for components in sing_comps]
        return self._singular_components

    
    @property
    def monodromy_matrices_morsification(self):
        if not hasattr(self, '_monodromy_matrices_morsification'):
            I1_monodromy_matrices = []
            for M in self.monodromy_matrices:
                type, base_change, nu = EllipticSingularities.monodromy_class(M)
                mats =  [base_change*M*base_change**-1 for M in EllipticSingularities.fibre_confluence[type][:-1]] + [base_change*EllipticSingularities.fibre_confluence[type][-1]*base_change**-1]*nu
                mats = [M.change_ring(ZZ) for M in mats]
                Mtot = 1
                for M2 in mats:
                    Mtot = M2*Mtot
                assert Mtot == M
                I1_monodromy_matrices += [mats]

            self._monodromy_matrices_morsification = I1_monodromy_matrices

        return self._monodromy_matrices_morsification

    @property
    def fiber(self) -> Hypersurface:
        if not hasattr(self,'_fiber'):
            self._fiber = Hypersurface(self.P(self.basepoint), nbits=self.ctx.nbits, fibration=self._fibration)
            if self._fiber.intersection_product == matrix([[0,-1], [1,0]]):
                self._fiber._extensions = list(reversed(self._fiber.extensions))
                del self._fiber._intersection_product
                del self._fiber._intersection_product_modification
            assert self.fiber.intersection_product == matrix([[0,1], [-1,0]])
        return self._fiber

    @property
    def thimbles(self):
        if not hasattr(self,'_thimbles'):
            self._thimbles=[]
            for pcs, path in zip(self.permuting_cycles, self.paths):
                for pc in pcs:
                    self._thimbles+=[(pc, path)]
        return self._thimbles

    @property
    def permuting_cycles(self):
        if not hasattr(self, '_permuting_cycles'):
            self._permuting_cycles = [[] for i in range(len(self.monodromy_matrices))]
            for i in range(len(self.monodromy_matrices)):
                M = self.monodromy_matrices[i]
                D, U, V = (M-1).smith_form()
                for j in range(2):
                    if D[j,j]!=0:
                        self._permuting_cycles[i] += [ V*vector([1 if k==j else 0 for k in range(2)]) ]
        return self._permuting_cycles
    
    @property
    def permuting_cycles_morsification(self):
        if not hasattr(self, '_permuting_cycles_morsification'):
            monodromy_matrices = flatten(self.monodromy_matrices_morsification)
            vanishing = flatten(self.vanishing_cycles_morsification)
            self._permuting_cycles_morsification = []
            for i in range(len(monodromy_matrices)):
                M = monodromy_matrices[i]
                D, U, V = (M-1).smith_form()
                p = V.column(0)
                if (M-1)*p != vanishing[i]:
                    p = -p
                assert (M-1)*p == vanishing[i]
                self._permuting_cycles_morsification += [ p ]
        return self._permuting_cycles_morsification

    @property
    def borders_of_thimbles(self):
        if not hasattr(self, '_borders_of_thimbles'):
            self._borders_of_thimbles = []
            for ps, M in zip(self.permuting_cycles, self.monodromy_matrices):
                self._borders_of_thimbles += [(M-1)*p for p in ps]
        return self._borders_of_thimbles


    @property
    def infinity_loops(self):
        """The linear combinations of thimbles that correspond to extensions along the (trivial) loop around infinity."""
        if not hasattr(self, '_infinity_loops'):
            infinity_cycles = []
            for i in range(2):
                v = vector([1 if k==i else 0 for k in range(2)])
                coefs = []
                for j in range(len(self.critical_values)):
                    M = self.monodromy_matrices[j]
                    if len(self.permuting_cycles[j])==0:
                        continue
                    coefs += list(matrix([(M-1)*t for t in self.permuting_cycles[j]]).solve_left((M-1)*v))
                    v = self.monodromy_matrices[j]*v
                infinity_cycles+=[vector(coefs)]
            self._infinity_loops = matrix(infinity_cycles).change_ring(ZZ).rows()
        return self._infinity_loops

    @property
    def extensions_morsification(self):
        """Representant of the extensions of the morsification of the elliptic surface. 
        Along with the fibre and section, this constitutes a basis for the second homology group of the surface. 
        The singular fibre components are identified at the end of the list."""
        if not hasattr(self, '_extensions_morsification'):
            singular_components = flatten(self.singular_components, max_level=2)
            infinity_loops = [self.morsify(v) for v in self.infinity_loops]
            delta = matrix(flatten(self.vanishing_cycles_morsification)).change_ring(ZZ)
            kerdelta = delta.kernel().matrix()
            D, U, V = kerdelta.smith_form()
            B = D.solve_left(matrix(infinity_loops)*V).change_ring(ZZ)*U
            quotient_basis = Util.find_complement(B)
            if quotient_basis.nrows()==0:
                self._extensions_morsification = kerdelta.submatrix(0,0,0).rows()
            else:
                self._extensions_morsification = (quotient_basis*kerdelta).rows()
        return self._extensions_morsification
    
    @property
    def extensions(self):
        """Representants of the extensions of the elliptic surface."""
        if not hasattr(self, '_extensions'):
            delta = matrix(self.borders_of_thimbles).change_ring(ZZ)
            kerdelta = delta.kernel().matrix()
            D, U, V = kerdelta.smith_form()
            B = D.solve_left(matrix(self.infinity_loops)*V).change_ring(ZZ)*U
            quotient_basis = Util.find_complement(B)
            if quotient_basis.nrows()==0:
                self._extensions = kerdelta.submatrix(0,0,0).rows()
            else:
                self._extensions = (quotient_basis*kerdelta).rows()
        return self._extensions

    @property
    def primary_lattice(self):
        if not hasattr(self, '_primary_lattice'):
            extensions = [self.lift(self.morsify(v)) for v in self.extensions]
            singular_components =[self.lift(v) for v in flatten(self.singular_components, max_level=2)]
            self._primary_lattice = extensions + singular_components + [self.fibre_class, self.section]
        return self._primary_lattice
   
    def lift(self, v):
        """Given a combination of thimbles of morsification, gives the corresponding homology class"""
        infinity_loops = [self.morsify(v) for v in self.infinity_loops]
        v = matrix(self.extensions_morsification + infinity_loops).solve_left(v)
        return vector(list(v)[:-len(self.infinity_loops)] + [0,0])

    @property
    def homology(self):
        if not hasattr(self, '_homology'):
            self._homology = identity_matrix(len(self.extensions_morsification)+2).rows()
        return self._homology


    @property
    def transition_matrices(self):
        if not hasattr(self, '_transition_matrices'):
            transition_matrices = []
            for L in self.picard_fuchs_equations:
                L = L* L.parent().gens()[0]
                transition_matrices += [self.integrate(L)]
                if "infinity" in self.critical_values:
                    transition_matrix_infinity = 1
                    for M in transition_matrices[-1]:
                        transition_matrix_infinity = M*transition_matrix_infinity
                    transition_matrices[-1] += [transition_matrix_infinity**(-1)]
            self._transition_matrices = transition_matrices
        return self._transition_matrices
    
    @property
    def cyclic_form(self):
        if not hasattr(self, '_cyclic_form'):
            for v in [[1,0], [0,1], [1,1]]:
                L = self._family.picard_fuchs_equation(vector(v))
                if L.order() == 2:
                    break
            assert L.order() == 2, "could not find cyclic  Picard-Fuchs equation"
            self._cyclic_form = v
            self._cyclic_picard_fuchs_equation = L
        return self._cyclic_form
    
    @property
    def cyclic_picard_fuchs_equation(self):
        if not hasattr(self, '_cyclic_picard_fuchs_equation'):
            self.cyclic_form
        return self._cyclic_picard_fuchs_equation
    
    @property
    def cyclic_transition_matrices(self):
        if not hasattr(self, '_cyclic_transition_matrices'):
            L = self.cyclic_picard_fuchs_equation
            L = L* L.parent().gens()[0]
            self._cyclic_transition_matrices = self.integrate(L)
        return self._cyclic_transition_matrices
    
    def integrate(self, L):
        logger.info("Computing numerical transition matrices of operator of order %d and degree %d (%d edges total)."% (L.order(), L.degree(), len(self.fundamental_group.edges)))
        begin = time.time()

        integrator = Integrator(self.fundamental_group, L, self.ctx.nbits)
        transition_matrices = integrator.transition_matrices
        
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("Integration finished -- total time: %s."% (duration_str))

        return transition_matrices

    def forget_transition_matrices(self):
        del self._transition_matrices
        del self._integrated_thimbles
        
    @property
    def integrated_thimbles(self):
        if not hasattr(self, '_integrated_thimbles'):
        
            s=len(self.fiber.extensions)
            r=len(self.thimbles)
            
            pM = self.fiber.period_matrix
            integration_correction = diagonal_matrix([1/ZZ(factorial(k)) for k in range(s+1)])

            _integrated_thimbles_all=[]
            for transition_matrices, w in zip(self.transition_matrices, self.holomorphic_forms):
                derivatives_at_basepoint = self.derivatives_values_at_basepoint(w)
                cohomology_fiber_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fiber.cohomology], self.basepoint)
                
                initial_conditions = integration_correction * derivatives_at_basepoint * cohomology_fiber_to_family.inverse() * pM
                _integrated_thimbles = []
                for i,ps in enumerate(self.permuting_cycles):
                    _integrated_thimbles += [(transition_matrices[i]*initial_conditions*p)[0] for p in ps]
                _integrated_thimbles_all += [_integrated_thimbles]
            self._integrated_thimbles = _integrated_thimbles_all
        return self._integrated_thimbles


    def derivatives_values_at_basepoint(self, w):
        s=len(self.fiber.extensions)

        derivatives = [self.P.parent()(0), w]
        for k in range(s-1):
            derivatives += [self._derivative(derivatives[-1], self.P)] 
        return self.family._coordinates(derivatives, self.basepoint)

    def _compute_intersection_product(self):
        r=len(flatten(self.vanishing_cycles_morsification))
        extensions = matrix(self.extensions_morsification)
        inter_prod_thimbles = matrix([[self._compute_intersection_product_thimbles(i,j) for j in range(r)] for i in range(r)])
        intersection_11 = (-1) * (extensions*inter_prod_thimbles*extensions.transpose()).change_ring(ZZ)
        intersection_02 = matrix(ZZ, [[0,1],[1,-2]])
        return block_diagonal_matrix(intersection_11, intersection_02)
        
    def _compute_intersection_product_thimbles(self,i,j):
        vi = self.permuting_cycles_morsification[i]
        Mi = flatten(self.monodromy_matrices_morsification)[i]
        vj = self.permuting_cycles_morsification[j]
        Mj = flatten(self.monodromy_matrices_morsification)[j]

        di, dj = ((Mi-1)*vi), (Mj-1)*vj

        
        res = di*self.fiber.intersection_product*dj
        resid = -vi*self.fiber.intersection_product*di

        if i==j:
            return resid
        if i<j:
            return res
        else:
            return 0

    @classmethod
    def _derivative(self, A, P): 
        """computes the numerator of the derivative of A/P^k"""
        field = P.parent().fraction_field()
        return field(A).derivative() - A*P.derivative()         

    @property
    def fundamental_group(self):
        if not hasattr(self,'_fundamental_group'):
            begin = time.time()

            fundamental_group = FundamentalGroupVoronoi(self.critical_values, self.basepoint) # access future delaunay implem here
            fundamental_group.sort_loops()

            end = time.time()
            duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
            logger.info("Fundamental group computed in %s."% (duration_str))

            self._critical_values = fundamental_group.points[1:]
            self._fundamental_group = fundamental_group
        return self._fundamental_group

    @property
    def paths(self):
        if not hasattr(self,'_paths'):
            paths = []
            for path in self.fundamental_group.pointed_loops:
                paths += [[self.fundamental_group.vertices[v] for v in path]]
            self._paths= paths
        return self._paths

    @property
    def basepoint(self):
        if  not hasattr(self, '_basepoint'):
            shift = 1
            reals = [self.ctx.CF(c).real() for c in self.critical_values]
            xmin, xmax = min(reals), max(reals)
            self._basepoint = Util.simple_rational(xmin - (xmax-xmin)*shift, (xmax-xmin)/10)
        return self._basepoint

    @property
    def neron_severi(self):
        if  not hasattr(self, '_neron_severi'):
            self._neron_severi = IntegerRelations(self.periods_matrix.transpose()).basis.rows()
        return self._neron_severi
    
    @property
    def transcendental_lattice(self):
        if  not hasattr(self, '_transcendental_lattice'):
            IL = IntegralLattice(self.intersection_product)
            self._transcendental_lattice = IL.orthogonal_complement(self.neron_severi).basis()
        return self._transcendental_lattice

    @property
    def trivial_lattice(self):
        if  not hasattr(self, '_trivial_lattice'):
            singular_components = flatten(self.singular_components)
            self._trivial_lattice = [self.lift(v) for v in singular_components] + [self.fibre_class, self.section]
        return self._trivial_lattice
    
    @property
    def fibre_class(self):
        return self.homology[-2]
    @property
    def section(self):
        return self.homology[-1]
    
    @property
    def mordell_weil(self):
        if  not hasattr(self, '_mordell_weil'):
            NS = matrix(self.neron_severi).image()
            Triv = NS.submodule(self.trivial_lattice)
            self._mordell_weil = NS/Triv
        return self._mordell_weil
    
    @property
    def essential_lattice(self):
        if  not hasattr(self, '_essential_lattice'):
            IL = IntegralLattice(self.intersection_product)
            self._essential_lattice = IL.sublattice(self.neron_severi).orthogonal_complement(self.trivial_lattice).basis_matrix()
        return self._essential_lattice
    
    @property
    def mordell_weil_lattice(self):
        ess_coords = matrix(self.neron_severi).solve_left(self.essential_lattice).change_ring(ZZ)
        triv_coords = matrix(self.neron_severi).solve_left(matrix(self.trivial_lattice)).change_ring(ZZ)
        coords = block_matrix([[ess_coords],[triv_coords]])
        projection_temp = block_diagonal_matrix([identity_matrix(len(self.essential_lattice.rows())), zero_matrix(len(self.trivial_lattice))])
        orth_proj = coords**-1*projection_temp*coords

        quotient_basis = Util.find_complement(triv_coords)
        coordsMW = quotient_basis*orth_proj
        A = coordsMW*matrix(self.neron_severi)
        return A*self.intersection_product*A.transpose()

    @property
    def types(self):
        if not hasattr(self, "_types"):
            self._types = [EllipticSingularities.monodromy_class(M) for M in self.monodromy_matrices]
        return self._types

    @property
    def holomorphic_forms(self):
        if not hasattr(self, "_holomorphic_forms"):
            L = self._family.picard_fuchs_equation(vector([1,0]))
            t = L.base_ring().gens()[0]

            bs = []
            rs = []
            roots = L.leading_coefficient().roots(QQbar, multiplicities=False)
            for r in roots:
                if r in self.critical_values:
                    i = self.critical_values.index(r)
                    ty, _, n = self.types[i]
                    bs += [self._b(L, r, ty, n)]
                else:
                    bs += [self._b(L, r, "I", 0)]
                rs += [self._rnorm(L, r)]

            L2 = L.annihilator_of_composition(1/t)
            if L2.leading_coefficient()(0) ==0:
                r = "infinity"
                roots += [r]
                if r in self.critical_values:
                    i = self.critical_values.index(r)
                    ty, _, n = self.types[i]
                    bs += [self._b(L2, 0, ty, n)]
                else:
                    bs += [self._b(L2, 0, "I", 0)]
                rs += [self._rnorm(L2, 0)]
            ords = [0  if r!="infinity" else -2*2 for r in roots]
            
            div = - vector(rs) - vector(bs) + vector(ords)

            print(div)

            Z = 1
            pols = []
            degrees = []
            for i, r in enumerate(roots):
                if r=="infinity":
                    continue
                pol = r.minpoly()
                if pol in pols:
                    assert degrees[pols.index(pol)] == div[i]
                    continue
                pols += [pol]
                degrees += [div[i]]
                Z = Z*pol**(-div[i])
            Z = Z(t)

            Dt = L.parent().gens()[0]
            W = (Dt + L.coefficients()[1]/L.coefficients()[2]).rational_solutions()[0][0]

            # assert sum(div) ==0, "Multiple or no holomorphic forms, not implemented yet"

            self._holomorphic_forms = [-Z/W*t**i for i in range(0, sum(div)+1)]
        return self._holomorphic_forms
    
    @classmethod
    def _rnorm(cls, L, p):
        R = L.base_ring()
        t = R.gens()[0]
        return floor((L.local_basis_monomials(p)[0]**12)(t=t).degree(t)/12)
    
    @classmethod
    def _b(cls, L, p, ty, n):
        if ty=="I" and n>0:
            return -1
        R = L.base_ring()
        t = R.gens()[0]
        m = L.local_basis_monomials(p)[1] / L.local_basis_monomials(p)[0]
        m=m(t=t+p)
        if ty=="I":
            l = (R(m).degree()-2)
            if l==-1:
                return 0
            else:
                return l+1
        if ty=="II":
            return (R(m**6).degree()-2)/6
        if ty=="III":
            return (R(m**4).degree()-2)/4
        if ty=="IV":
            return (R(m**3).degree()-2)/3
        if ty=="I*":
            if n==0:
                return (R(m**2).degree()-2)/2
            else:
                return -1
        if ty=="II*":
            return (R(m**6).degree()-4)/6 - 1
        if ty=="III*":
            return (R(m**4).degree()-2)/4 - 1
        if ty=="IV*":
            return (R(m**3).degree()-1)/3 - 1


# -*- coding: utf-8 -*-

import sage.all

from sage.rings.complex_mpfr import ComplexField
from sage.graphs.graph import Graph
from sage.symbolic.constants import I
from sage.functions.other import arg

from sage.misc.flatten import flatten

class FundamentalGroupDelaunay(object):
    def __init__(self, points, basepoint):
        assert basepoint not in points

        self._points = [basepoint] + points

        self.CC = ComplexField(500)


    @property
    def points(self):
        return self._points

    @property
    def basepoint(self):
        return self.points[0]

    @property
    def npoints(self):
        return len(self.points)
    

    @property
    def minimal_graph(self):
        if not hasattr(self, '_minimal_graph'):
            tree = Graph(self.npoints) 
            edges = flatten([[(i,j) for i in range(j)] for j in range(self.npoints)], max_level=1) 
            edges.sort(key=(lambda e: abs(self.CC(self.points[e[0]])-self.CC(self.points[e[1]])))) # we sort edges by length
            for e in edges:
                if len(tree.shortest_path(e[0], e[1]))==0:
                    tree.add_edge(e)
            self._minimal_graph = tree

        return self._minimal_graph

    
    @property
    def neighbours(self):
        if not hasattr(self, '_neighbours'):
            smallangleshift = -I/100000000000 # this assumes there are no angle very close to -pi, so we can send angles equal to -pi below the line
            neighbours=[[] for vertex in self.minimal_graph.vertices()]
            for vertex in self.minimal_graph.vertices():
                l = self.minimal_graph.neighbors(vertex)
                l.sort(key=lambda v2:arg(self.CC(self.points[v2] - self.points[vertex] + smallangleshift/abs(self.CC(self.points[v2] - self.points[vertex])))))
                neighbours[vertex] = l
            self._neighbours = neighbours
        return self._neighbours

    def _visit_neighbours(self, vertex, parent):
        neighbours = self.neighbours[vertex]
        if len(neighbours) == 1:
            return [[vertex]]
        assert parent in neighbours, "parent node is not a neighbour"

        prefix=[vertex]
        index_parent = neighbours.index(parent)

        i=index_parent+1
        if i==len(neighbours):
            i=0
            prefix=[vertex, "loop"]
        paths=[]
        while i != index_parent:
            paths = paths+[prefix+path_child for path_child in self._visit_neighbours(neighbours[i], vertex)]
            i=i+1
            if i == len(neighbours):
                i=0
                prefix=[vertex, "loop"]
        return paths + [[vertex]]

    @property
    def paths(self):
        if not hasattr(self, "_paths"):
            paths = []
            for vertex in self.neighbours[0]:
                paths= paths + self._visit_neighbours(vertex, 0)
            paths = [[0]+path for path in paths]
            self._paths = paths
        return self._paths
    






















































from sage.arith.misc import random_prime
from sage.functions.other import floor
from sage.misc.cachefunc import cached_method
from sage.rings.finite_rings.finite_field_constructor import FiniteField
from sage.rings.polynomial.polynomial_element import *
from sage.rings.polynomial.polynomial_ring import *
from sage.matrix.constructor import Matrix
from sage.modules.free_module_element import vector
from sage.rings.rational_field import QQ
from sage.rings.integer_ring import ZZ

from ore_algebra.ore_algebra import DifferentialOperators
from ore_algebra import ore_operator

import logging

def subproduct_tree(self, points):
    """Return the the subproduct tree associated to the points. The root is the
    polynomial `prod(t - p for p in points)` and the left and right children are
    the subproduct trees associated to the first and second half of the list
    `points`, respectively.

    """
    if len(points) <= 0:
        return None
    if len(points) == 1:
        return (self.gen() - points[0], )

    left = subproduct_tree(self, points[:len(points)//2])
    right = subproduct_tree(self, points[len(points)//2:])

    return (left[0]*right[0], left, right)

PolynomialRing_commutative.subproduct_tree = subproduct_tree


def polynomial_multi_evaluation_with_spt(self, spt, ret=None):
    if spt == None:
        return []

    if ret is None:
        ret = []

    rem = self.mod(spt[0])
    if len(spt) == 1:
        ret.append(rem.constant_coefficient())
    else:
        polynomial_multi_evaluation_with_spt(rem, spt[1], ret)
        polynomial_multi_evaluation_with_spt(rem, spt[2], ret)

    return ret


def sum_fractions_with_spt(coeffs, spt):
    if len(spt) == 1:
        return spt[0].parent()(coeffs[0])
    else:
        left = sum_fractions_with_spt(coeffs[:len(coeffs)//2], spt[1])
        right = sum_fractions_with_spt(coeffs[len(coeffs)//2:], spt[2])
        return left*spt[2][0] + right*spt[1][0]



class EvaluationInterpolation:
    """Perform basic evaluation and interpolation operations."""
    def __init__(self, polring, points):
        """Args:
        - polring: a univariate polynomial ring
        - points: a list of distinct elements of the basering
        """
        self.points = points
        self.polring = polring
        self.spt = None

    def _compute_spt(self):
        if self.spt is None:
            self.spt = self.polring.subproduct_tree(self.points)

    def evaluate(self, pol):
        """Return the list of values of `pol` at the elements of `self.points`."""
        self._compute_spt()
        return polynomial_multi_evaluation_with_spt(pol, self.spt)

    def interpolate(self, values):
        """Return a polynomial which evaluate to `values[i]` at `self.points[i]`."""
        assert len(values) == len(self.points), "The number of values must match the number of points."
        self._compute_spt()
        A = self.spt[0]
        der = self.evaluate(A.derivative())
        if any(d == 0 for d in der):
            return None

        br = A.base_ring()
        coeffs = [br(values[i])/der[i] for i in range(len(der))]
        return sum_fractions_with_spt(coeffs, self.spt)

    def _rational_interpolate_gen(self, values):
        """Returns a pair of polynomials (n,d) such that n/d interpolates the pairs.
        May fail and return None.
        """
        if len(values) == 0:
            return self.zero(), self.one()

        self._compute_spt()
        A = self.spt[0]
        B = self.interpolate(values)
        try:
            numer, denom = B.rational_reconstruct(A)
            if numer.degree() + denom.degree() + 1 < len(values):
                return numer, denom
            else:
                return None
        except:                 # rational_reconstruct fails with an exception.
            return None

    def _rational_interpolate_qq(self, values):
        def ev(prime):
            gf = FiniteField(prime)
            modpolring = self.polring.change_ring(gf)
            ei = EvaluationInterpolation(modpolring, self.points)
            return ei._rational_interpolate_gen(values)

        return ModularReconstruction(ev).recons()

    def rational_interpolate(self, values):
        if self.polring.base_ring().is_subring(QQ) and len(self.points) > 8:
            return self._rational_interpolate_qq(values)
        else:
            return self._rational_interpolate_gen(values)


def interpolation(self, pairs):
    """Same as `lagrange_polynomial` but faster."""
    if len(pairs) == 0:
        return self.zero()

    points = [p[0] for p in pairs]
    evs = [p[1] for p in pairs]
    return EvaluationInterpolation(self, points).interpolate(evs)

PolynomialRing_field.interpolation = interpolation


def rational_interpolation(self, pairs):
    """Returns a pair of polynomials (n,d) such that n/d interpolates the pairs.
    May fail and return None.
    """
    if len(pairs) == 0:
        return self.zero(), self.one()

    points = [p[0] for p in pairs]
    evs = [p[1] for p in pairs]
    return EvaluationInterpolation(self, points).rational_interpolate(evs)

PolynomialRing_field.rational_interpolation = rational_interpolation



class Serial:
    def __init__(self, polring=None):
        self.polring = polring
        if not polring is None:
            self.dopring = DifferentialOperators(polring.base_ring(), var=polring.variable_name())

        self.polrings = {}

    def _explode(self, elt, data, struct):
        if isinstance(elt, list):
            struct.append("list")
            struct.append(len(elt))
            for e in elt:
                self._explode(e, data, struct)
        elif isinstance(elt, tuple):
            struct.append("tuple")
            self._explode(list(elt), data, struct)
        elif isinstance(elt, dict):
            struct.append("dict")
            struct.append(tuple(elt.keys()))
            self._explode(list(elt.values()), data, struct)
        elif isinstance(elt, sage.structure.element.Matrix):
            struct.append("matrix")
            self._explode(elt.rows(), data, struct)
        elif isinstance(elt, sage.structure.element.Vector):
            struct.append("vector")
            self._explode(elt.list(), data, struct)
        elif isinstance(elt, sage.rings.polynomial.polynomial_element.Polynomial):
            struct.append("polynomial")
            struct.append(elt.parent().variable_name())
            self._explode(elt.list(), data, struct)
        elif isinstance(elt, sage.rings.fraction_field_element.FractionFieldElement_1poly_field):
            struct.append("ratfun")
            self._explode(elt.numerator(), data, struct)
            self._explode(elt.denominator(), data, struct)
        elif isinstance(elt, ore_operator.UnivariateOreOperator):
            struct.append("dop")
            struct.append(elt.parent().base_ring().variable_name())
            self._explode(elt.list(), data, struct)
        else:
            struct.append("self")
            data.append(elt)

    def explode(self, elt):
        data = []
        struct = []
        self._explode(elt, data, struct)
        data.reverse()
        struct.reverse()
        return (data, tuple(struct))

    def recons(self, data, struct):
        return self._recons(data, list(struct))

    def _recons(self, data, struct):
        next = struct.pop()
        if next == "list":
            return [self._recons(data, struct) for i in range(struct.pop())]
        elif next == "tuple":
            return tuple(self._recons(data, struct))
        elif next == "dict":
            keys = struct.pop()
            values = self._recons(data, struct)
            return dict(zip(keys, values))
        elif next == "matrix":
            return Matrix(self._recons(data, struct))
        elif next == "vector":
            return vector(self._recons(data, struct))
        elif next == "polynomial":
            name = struct.pop()
            l = self._recons(data, struct)
            if len(l) == 0:
                return 0
            else:
                if not name in self.polrings:
                    self.polrings[name] = PolynomialRing(l[0].parent(), name)
                return self.polrings[name](l)
        elif next == "ratfun":
            numer = self._recons(data, struct)
            denom = self._recons(data, struct)
            return numer/denom
        elif next == "dop":
            name = struct.pop()
            l = self._recons(data, struct)
            if len(l) == 0:
                return 0
            else:
                Dop, _, _ = DifferentialOperators(l[0].parent().base_ring(), var=name)
                return Dop(l)
        elif next == "self":
            return data.pop()
        else:
            raise Exception("Misformed struct")


class Tick:
    def __init__(self, inc=4):
        self.cnt = 1
        self.i = 1
        self.inc = inc

    def tick(self):
        self.i -= 1
        if self.i == 0:
            self.cnt = self.cnt+self.inc
            self.i = self.cnt
            return True
        else:
            return False

    def ticknexttime(self):
        self.i = 1

class ModularReconstruction:
    logger = logging.getLogger('numperiods.interpolation.ModularReconstruction')

    def __init__(self, evaluator, modsize=30):
        self.maxprime = 2**modsize
        self.data = {}
        self.primes = {}
        self.cand0 = None
        self.evaluator = evaluator
        self.serial = Serial()
        self.tick = Tick(inc=1)

    def _next(self, prime):
        if prime in self.primes:
            return

        self.logger.info("Evaluating modulo %i" % prime)

        try:
            ev = self.evaluator(prime)
        except ZeroDivisionError:
            self.logger.info("Bad evaluation, skipping this value.")
            return

        data, struct = self.serial.explode(ev)
        key = struct

        if not key in self.data:
            self.data[key] = data
        else:
            assert len(data) == len(self.data[key]), "Evaluations must have the same length for a given key."
            self.data[key] = [self.data[key][i].crt(data[i]) for i in range(len(data))]

        return key

    def _try_reconstruction(self, key):
        self.logger.info("Trying modular reconstruction.")

        cand = []
        try:
            for c in self.data[key]:
                cand.append(c.rational_reconstruction())
        except ArithmeticError:
            return None

        if cand == self.cand0:
            return self.serial.recons(cand, key)
        else:
            self.logger.info("Possible reconstruction. Computing modulo one more prime to check.")
            self.tick.ticknexttime()
            self.cand0 = cand
            return None

    def recons(self):
        pt = Integer(1)
        while True:
            pt += 1
            prime = random_prime(self.maxprime, lbound=self.maxprime/128)
            key = self._next(prime)

            # We don't always try reconstruction (it is expensive)
            if self.tick.tick():
                cand = self._try_reconstruction(key)
                if not cand is None:
                    return cand


class FunctionReconstruction:
    logger = logging.getLogger('numperiods.interpolation.FunctionReconstruction')

    def __init__(self, polring, evaluator):
        self.polring = polring
        self.serial = Serial()
        self.evaluator = evaluator

        self.tick = Tick()
        self.data = {}
        self.tests = {}
        self.testsmod = {}
        self.rands = {}
        self.basering = polring.base_ring()
        if self.basering.characteristic() == 0:
            self.modring = FiniteField(random_prime(10**9))
            self.modpolring = polring.change_ring(self.modring)
        else:
            self.modring = self.basering
            self.modpolring = polring

    def _next(self, pt):
        self.logger.info("Evaluating at %i" % pt)
        try:
            ev = self.evaluator(pt)
        except ZeroDivisionError:
            self.logger.info("Bad evaluation, skipping this value.")
            return

        data, struct = self.serial.explode(ev)
        key = struct

        if not key in self.rands:
            self.data[key] = {}
            self.tests[key] = {}
            self.testsmod[key] = {}
            self.rands[key] = [ZZ.random_element(1000) for _ in data]
        else:
            assert len(data) == len(self.rands[key]), "Evaluations must have the same length for a given key."

        self.data[key][pt] = data

        try:
            self.tests[key][pt] = sum(r*e for r, e in zip(self.rands[key], data))
            self.testsmod[key][self.modring(pt)] = self.modring(self.tests[key][pt])
        except ZeroDivisionError:
            self.logger.info("Bad evaluation to the finite field, skipping this value.")
            return

        return key

    def recons(self, denomapart=False):
        pt = Integer(100)
        while True:
            pt += 1
            key = self._next(pt)

            # We don't always try reconstruction (it is expensive)
            if self.tick.tick():
                cand = self._try_reconstruction(key, denomapart=denomapart)
                if not cand is None:
                    return cand


    def _try_reconstruction(self, key, denomapart=False):
        self.logger.info("Trying rational reconstruction...")

        reconmod = self.modpolring.rational_interpolation(self.testsmod[key].items())
        if reconmod is None:
            self.logger.info("Reconstruction failed.")
            return None

        self.logger.info("Reconstructing denominator...")

        points = list(self.tests[key].keys())
        ei = EvaluationInterpolation(self.polring, points)
        if self.modring == self.basering:
            recon = reconmod
        else:
            recon = ei.rational_interpolate([self.tests[key][p] for p in points])
            if recon is None:
                self.logger.info("Reconstruction failed surprisingly.")
                return None

        denom = recon[1]
        assert denom.leading_coefficient() == 1

        evdenom = ei.evaluate(denom)
        nfun = len(self.rands[key])
        cand = []

        for i in range(nfun):
            self.logger.info("Reconstructing numerator %d of %d..." % (i+1, nfun))
            elt = ei.interpolate([self.data[key][p][i]*evdenom[idx] for idx, p in enumerate(points)])
            if 3*elt.degree() > 2*len(points):
                self.logger.warn("The random sampling failed. Should happen very rarely.")
                self.__init__(self.polring, self.evaluator)
                return None
            if denomapart:
                cand.append(elt)
            else:
                cand.append(elt/denom)

        if denomapart:
            return self.serial.recons(cand, key), denom
        else:
            return self.serial.recons(cand, key)





import logging

from sage.matrix.constructor import Matrix
from sage.rings.complex_arb import ComplexBallField
from sage.rings.complex_interval_field import ComplexIntervalField_class
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.real_arb import RealBallField
from sage.rings.real_mpfi import RealIntervalField_class, RealIntervalField
from sage.rings.real_mpfr import RR
from sage.rings.real_double import RDF

logger = logging.getLogger(__name__)

class PrecisionError(Exception):
    pass

class IntegerRelations(object):
    def __init__(self, mat, beta=None, threshold=None, min_threshold=10**3,
                 algorithm='fpLLL:proved', fp='xd', **kwd):
        """`mat` is a matrix with real/complex ball/interval coefficients. This class
        computes the lattice of integer relations between the rows of mat.

        """

        logger.info("Computing lattice of integer relations (ambient dimension %d, number of relations %d)."
                    % (mat.nrows(), mat.ncols()))

        m = mat.nrows()
        nrels = mat.ncols()
        if isinstance(mat.base_ring(), ComplexIntervalField_class) or isinstance(mat.base_ring(), ComplexBallField):
            real = mat.apply_map(lambda x: x.real())
            imag = mat.apply_map(lambda x: x.imag())
            self._imat = Matrix.block([[real, imag]])
            nrels *= 2
        else:
            self._imat = mat

        if not isinstance(self._imat.base_ring(), RealIntervalField_class):
            if isinstance(self._imat.base_ring(), RealBallField):
                self._imat = self._imat.change_ring(RealIntervalField(self._imat.base_ring().precision()))
            else:
                raise ValueError

        if not beta is None:
            self._beta = beta
        else:
            self._beta = (1/max(3*x.absolute_diameter() for x in self._imat.list())).floor()

        logger.info("log(beta) = %s" % str(RR(self._beta).log(10)))
        assert self._beta > 100

        if threshold is None:
            threshold = RR(self._beta)**(1/QQ(m))  # This should be quite conservative.
            logger.info("log(threshold) = %s", RR(threshold).log(10))
            if threshold < min_threshold:
                raise PrecisionError
        else:
            logger.info("log(threshold) = %s (forced by caller)", RR(threshold).log(10))


        self._imat = (self._beta*self._imat).apply_map(lambda x : x.center().round())

        self._rawlat = Matrix.block([[self._imat, Matrix.identity(ZZ, m)]])

        logger.info("Running LLL...")
        self._redlat = self._rawlat.LLL(delta=0.75, algorithm=algorithm, fp=fp, **kwd)
        self._norms = [r.norm() for r in self._redlat.change_ring(RR).rows()]

        self.rank = len([n for n in self._norms if n < threshold])
        kappa = RR(2)**(-QQ(m+1)/2)/m

        if self.rank == len(self._norms):
            self._threshold = RR('+Inf')
        else:
            self._threshold = kappa*self._norms[self.rank]
            #assert self.rank == 0 or self._norms[self.rank-1] < self._threshold

        self._expected_threshold = kappa*RR(self._beta)**(QQ(nrels)/(m - self.rank))

        # This will be ultimately exported to json, so we use primitive types.
        self._numdata = {
            "logB": int(RR(self._threshold.log(10)))
        }

        if self.rank > 0:
            self._numdata['N'] = int(self._norms[self.rank-1].ceil())
            self._numdata['logE'] = int( (self._norms[self.rank-1]/self._beta*m).log(10).ceil() )
            self._numdata['rarity'] = int(RR(self._beta).log(10)*(QQ(nrels)/(m - self.rank))) - RDF(1+self._numdata['N']).log(10).ceil()
        else:
            self._numdata['N'] = int(0)
            self._numdata['e'] = int(0)
            self._numdata['rarity'] = int(0)

        self.basis = self._redlat[:self.rank, nrels:]

    def basis_elements(self):
        self.basis_matrix().rows()

    def threshold(self):
        return self._threshold

    def half_certificate(self):
        return self._numdata

    # TODO Add info on the numeric reliability







# -*- coding: utf-8 -*-

from sage.rings.all import *
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.modules.free_module_element import vector
from sage.matrix.constructor import Matrix
from sage.rings.polynomial.polynomial_ring import *

from ..exceptions import NotSmoothError

class Cohomology(object):
    def __init__(self, f, shift=0, basisfor = None):
        """f, a homogeneous polynomial defining a smooth hypersurface in P^n.

        This class aims at computing in the n-th algebraic de Rham cohomology
        space of P^n - V(f), denoted H^n. Elements of H^n are represented as
        polynomials. A homogeneous polynomial p of degree s*deg(f)-n-1
        represents the differential form (s-1)! * p dx0...dxn / f^s (that is a
        degree 0 (n+1)-form on A^(n+1)-V(f) which induces a n-form on P^n-V(f).
        """

        assert f.is_homogeneous()

        self.nvars = f.parent().ngens()
        self.dim = self.nvars-2
        self.degree = f.degree()
        self.pol = f
        self.shift = shift
        self.R = f.parent()  # Variables x0...xn
        # Variables T x0...xn _x0..._xn
        self._R = PolynomialRing(f.parent().base_ring(), 2*self.nvars+1, ['T'] + [str(v) for v in self.R.gens()] + ['X'+str(v) for v in self.R.gens()], order='degrevlex')
        self._Tvar = self._R.gen(0)
        self._vars = [self._R.gen(i) for i in range(1, self.nvars+1)]
        self._xvars = {self._R.gen(i): self._R.gen(self.nvars+i) for i in range(1, self.nvars+1)}
        self._conv = self.R.hom([self._R.gen(i+1) for i in range(self.nvars)])
        self._iconv = self._R.hom([1] + list(self.R.gens()) + [0 for i in range(self.nvars)])
        self._pol = self._conv(f)

        self._jac = ideal([self._Tvar*self._pol.derivative(v) - self._xvars[v] for v in self._vars] + [v1*v2 for v1 in self._xvars.values() for v2 in self._xvars.values()])


            # We could do the computation in R in a simpler way, but doing it in _R ensures consistency.
        jac = self._pol.jacobian_ideal() + ideal(self._Tvar) + ideal(list(self._xvars.values()))
        if basisfor is None:
            if not f.jacobian_ideal().dimension() == 0:
                raise NotSmoothError()
            self.__basis = sorted([p*self._Tvar for p in jac.normal_basis() if (p.degree() + self.nvars + self.shift) % self.degree == 0])
            self._basis = [self._iconv(p) for p in self.__basis]
            self._basis_indices = {elt: idx for idx, elt in enumerate(self._basis)}
        else:
            f, g = basisfor
            _f = self._conv(f)


    def basis(self):
        """Return a basis of the n-th algebraic de Rham cohomology space of P^n - V(f).
        """
        return self._basis

    def holomorphic_forms(self):
        return [b for b in self._basis if b.degree() + self.dim + 2 + self.shift == self.degree]

    def index_of_basis_elt(self, b):
        return self._basis_indices[b]

    # TODO Add a function for simulatenous reduction
    def _red(self, p):
        redp = self._jac.reduce(p)
        if redp == p:
            return p
        else:
            c0 = redp.coefficient({self._Tvar:1})
            c1 = sum([redp.coefficient({self._xvars[v]:1}).derivative(v) for v in self._vars])
            return self._Tvar*c0 + self._red(self._Tvar*c1)

    def red(self, p):
        """Return the canonical form of p in H^n"""
        return self._iconv(self._red(self._Tvar*self._conv(p)))

    def _integrate(self, p):
        if p==0:
            return []
        else:
            redp = self._jac.reduce(p)
            c0 = redp.coefficient({self._Tvar:1})
            coefs = [redp.coefficient({self._xvars[v]:1}) for v in self._vars]
            c1 = sum([redp.coefficient({self._xvars[v]:1}).derivative(v) for v in self._vars])
            assert c0==0
            return [coefs] + self._integrate(self._Tvar*c1)

    def integrate(self, p):
        """Return w in H^n such that p = dw"""
        return [[self._iconv(v) for v in l] for l in self._integrate(self._Tvar*self._conv(p))]

    def _coordinates(self, p):
        red = self._red(p)
        return vector(red.monomial_coefficient(m) for m in self.__basis)

    def coordinates(self, p):
        """Return the coordinates of the canonical form of p in H^n, in the basis self.basis()."""
        return self._coordinates(self._red(self._Tvar*self._conv(p)))

    def multmat(self, p):
        conv = self._conv(p)
        return Matrix([ self._coordinates(self._red(conv*b)) for b in self.__basis ])


    def weight(self, b):
        deg = b.degree() + self.dim + 2 + self.shift
        assert deg % self.degree == 0
        return deg / self.degree

